/*
 *  SparkFun Demo Game
 *  Copyright (C) 2008  Embedded Engineering, LLC
 *  Game/Programming: Clay Cowgill
 *  Audio/FX:         Alec Bourque
 *  Version 1.0
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Uzebox is a reserved trade mark
*/

/*

About this program:
-------------------

This program demonstrates the new sprites engine which also features full screen scrolling.
The video memory map is arranged as a 32x32 tiles area in which you slide a 22x26 window. 
Up to 31 simultaneous sprites can be displayed on screen at once. However no more 
than 5 sprites can be visible on each scanline due to CPU limitation. Note that in this version, 
sprite #0 must point to a transparent tile. Sprites with lower indexes have higher priority. 
I.e: sprite #1 has higher priority than sprite #5. If more than five sprites crosses the 
same scanline 'sprites overflow' occurs. In this case, the sprites engine can be configured 
to support two options: 
	0=the lower priority one will be completely invisible. 
	1=cycle the sprites so that all sprites on that scanline would flicker in turn (similar to the NES). 

Sprites are accessed using a structure with X,Y and tileIndex values. The sprites engine
uses a dedicated tile table that does not need to be the same as for the background tiles. 
This means that you can have a maximum of 256 tiles used for the sprites per rendering frame.

The engine now support screen sections (aka split screens) that replaces the overlay. Screen sections
are vertical section of the screen that are parametrizable at runtime. Parameter are:

	-scrollX: x displacement
	-scrollY: y displacement
	-height: section height in scanlines
	-vramBaseAdress: location in vram where this section starts rendering
	-tileTableAdress: tile set to use when rendering this section
	-wrapLine: Define at what Y scroll within that section Y-wrap will happen (Normally Y-Wrap happens at 0xff). So
	           instead of wrapping each 32 tiles you can wrap each 12 or 12 tiles+ 3 lines, etc.
	-flags: 1=sprites will be drawn on top of this section. 0=sprites will be under.
	
	Notes: 
	-The area of memory allocated for each section *must* be 32 tiles wide.
	-Number of sections is a compile option (see defines.h)
	-Total heights of all sections must be 208 (VRAM_TILES_V*TILE_HEIGHT). Any sections after line 208 is ignored		
	-Sections must be at least 1 line high except if they are at the end *and* not visible/active. So you can
	 not have zero height sections in the middle of section > 1 in height.

Due to the non-binary aligment of the tiles on the horizaontal axis (tile width=6), the caller needs to 
perform X wrapping before calling setting the screen section's scrollX variable. Wrapping happens 
at value X_SCROLL_WRAP (see uzeboxVideoEngine.c for details). Y Scrolling doesnt 
need this because it wraps automatically at 0xff. The sprites positions
are independant of the background scrolling posisition. To quickly turn off a 
sprite, set its Y position to -1 (or out of screen if you prefer). Finally,
it worth noting that horizontal scrolling was not trivial and the current method
requires around 9K of program space due to unrolled loops. If your games do not
need horizontal scrolling (i.e.: a "River Raid" game), you can specifiy a custom compilation option
to remove the code for this functionnality.

This revision also include and option to have music channel 3 play PCM sample instead of the noise channel.
Samples can be of arbitrary lenght and have a loop. Samples are defined as patches and can have a
command/modifier stream attached to them for envelopes and other effects. The patch system had to 
be rewritten to support PCM samples, so that may imply some adjustment on existing code.
This demo uses a PCM loop as background music.


Hope you have fun with it!


Uze


See Also:
---------
kernel/uzebox.h: API functions and variables
kernel/defines.h: Global defines, constants & compilation options
data/mypatches.h: PCM sample patch

*/

#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "kernel/uzebox.h"

// external graphics data (tiles for background and sprites)
#include "sparkfun_ts2.c"
#include "sparkfun_base.c"
#include "sparkfun_font.c"
#include "sparkfun_smoke.c"
#include "sparkfun_spr.c"
#include "sparkfun_sparks.c"
#include "sparkfun_zap.c"
// external audio data (patches and samples)
#include "data/patches.h"

// array of smoke maps to be drawn with DrawMap2
const char smoke_anim[6*7] PROGMEM=
{
2,2,
1,2,
15,16,
2,2,
3,4,
17,18,
2,2,
5,6,
19,20,
2,2,
7,8,
21,22,
2,2,
9,10,
23,24,
2,2,
11,12,
25,26,
2,2,
13,14,
27,28
};

#define sparkfun_sparkfun_map_Width   (20)
#define sparkfun_sparkfun_map_Height  (8)

// map of the SparkFun logo
const char sparkfun_sparkfun_map[] PROGMEM=
{sparkfun_sparkfun_map_Width,sparkfun_sparkfun_map_Height,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  3,  4,  1,  1,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  5,  6,  7,  8,  1,  1,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  9, 10, 11, 12,  1,  1,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1, 13, 14, 15, 16,  1,  1,  1,  1,  1,  1,  1,
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34,  1,  1,
   35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53,  1,
   54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73,
    1,  1, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91
};

#define finger_FingerMap_Width   (6)
#define finger_FingerMap_Height  (10)
#define fb 90

//and the finger map
const char finger_FingerMap[] PROGMEM=
{finger_FingerMap_Width,finger_FingerMap_Height,
       1,  fb+2,  fb+3,  fb+4,  fb+5,  fb+6,
    fb+7,  fb+8,  fb+9, fb+10, fb+11, fb+12,
   fb+13, fb+14, fb+15, fb+16, fb+17, fb+18,
   fb+19, fb+20, fb+21, fb+22, fb+23, fb+24,
   fb+25, fb+26, fb+27, fb+28, fb+29, fb+30,
   fb+31, fb+32, fb+33, fb+34, fb+35, fb+36,
   fb+37, fb+38, fb+39, fb+40, fb+41,     1,
   fb+42, fb+43,     1,     1,     1,     1,
   fb+44, fb+45,     1,     1,     1,     1,
   fb+46, fb+47,     1,     1,     1,     1
};

const char finger_sine[] PROGMEM=
{
  16,  17,  19,  20,  21,  23,  24,  25
,  26,  27,  28,  29,  30,  30,  31,  31
,  31,  31,  31,  30,  30,  29,  28,  28
,  27,  25,  24,  23,  22,  20,  19,  17
,  16,  14,  13,  11,  10,  8,  7,  6
,  5,  4,  3,  2,  1,  1,  0,  0
,  0,  0,  0,  1,  1,  2,  2,  3
,  4,  5,  7,  8,  9,  11,  12,  14

};

// little sine table for 'windy' smoke effect
const char windy[] PROGMEM=
{
3,5,6,7,6,5,3,2,1,0,1,2
}; // 12 elements

unsigned char x_locations[] PROGMEM=
	{0,18,42,60,84,102,120};

unsigned char hit_locations1[] PROGMEM=
	{0x08,0x18,0x2e,0x42,0x58,0x6c,0x7f};

unsigned char hit_locations2[] PROGMEM=
	{0x0d,0x1f,0x32,0x49,0x5c,0x73,0x84};

unsigned char hit_masks[] PROGMEM=
	{0x80,0x40,0x20,0x10,0x08,0x04,0x02};

unsigned char sx3=0;

// local variables are for wimps
unsigned char player_x, player_y;
unsigned char reload_time;
unsigned char bases=0xfe;
unsigned char bases_old=0xfe;
unsigned char sync_count=0;
unsigned char temp_left=0;
unsigned char temp_right=0;
unsigned char instructions=1; // display instructions =1
unsigned char game_start=0;
unsigned char wave=0;

// a couple structures for the bullets and enemies
struct missileStruct
{
	unsigned int x;
	unsigned int y;
	unsigned int destx;
	unsigned int desty;
	int xstep;
	int ystep;
	unsigned char range;
	unsigned char state;
};

struct sparkStruct
{
	unsigned int x;
	unsigned int y;
	int xstep;
	int ystep;
	unsigned char anim_timer;
	unsigned char state;
};

struct missileStruct missiles[8];  // up to eight missiles in the air
struct sparkStruct sparks[16];     // and up to sixteen enemies coming down

unsigned char exp_x; // explosion animation x location and state
char  exp_state=-1;  // component explosion status

unsigned char processControls(void);

// instructions to player/credits
const char strProtect[] PROGMEM ="PROTECT  COMPONENTS!"; 
const char strTitle[] PROGMEM  ="    E.S.D. ATTACK";
const char strRev[] PROGMEM    ="       VER 1.0";
const char strCopy[] PROGMEM   =" # 2008,CLAY COWGILL";
const char strGameover[] PROGMEM   ="PCB SMOKED! GAME OVER!";
const char strWave[] PROGMEM ="WAVE:";
const char strBlank[] PROGMEM ="                    ";
const char strHigh[] PROGMEM  ="   REACHED WAVE:      ";


//const char strScroller[] PROGMEM   ={0,1,2,3,4,5,6,32,33,34,35,36,37,38};
const char strScroller[] PROGMEM="GAME/PROGRAM: CLAY COWGILL * AUDIO: ALEC BOURQUE * MOVE CROSSHAIR WITH D-PAD AND FIRE YOUR MOVS AT THE SPARKS WITH A AND B BUTTONS.  FIRE TOO OFTEN AND YOUR MOVS WILL OVERHEAT AND MUST COOL DOWN...                                      ";

// if explosion is active, deal with it
void process_explosions()
{
	if (exp_state>=0) // active explosion
		{
		sprites[30].tileIndex=6+exp_state; // explosion uses two sprites, base address of the explosion
		sprites[31].tileIndex=7+exp_state; // graphics is tile 6 and 7
		exp_state+=2; // skip two at a time for animation
		}
	if (exp_state>18) // we over-run the explosion frames and go into the sparks, just because it looked neat
		{
		exp_state=-1; // explosion over, reset
		sprites[30].y=-1; // move explosion sprites off screen.  This is all sprites 30 and 31 are used for.
		sprites[31].y=-1;
		}
}

void start_explosion(unsigned char status)
{
	if (exp_state==-1) // if no other explosion is already in process
		{
		if (status==0x02) // figure out the explosion location, then set the x-coordinate for it
			exp_x=126;
		if (status==0x04)
			exp_x=108;
		if (status==0x08)
			exp_x=87;
		if (status==0x10)
			exp_x=66;
		if (status==0x20)
			exp_x=45;
		if (status==0x40)
			exp_x=24;
		if (status==0x80)
			exp_x=6;

		sprites[30].x=exp_x; // setup sprites locations and starting tiles for explosion graphic
		sprites[30].y=198;
		sprites[30].tileIndex=6;
		sprites[31].x=exp_x+6;
		sprites[31].y=198;
		sprites[31].tileIndex=7;

		TriggerNote(3,6,30,0xff); // exploding component soundfx 
		
		exp_state=0; // start the explosion (here a positive number means it's running)
		}
}

void update_base(unsigned char status)
{
	// manually calculated tile index numbers, ugly and hard to alter.  Not recommend to do this way.

	//mov blown = 90+48+0,1
	// blank space = 2
	//diode blown = 90+48+3,4
	//cap blown = 90+48+5
	//resistor blown = 90+48+6,7 
    // PCB FR4 = 90+48+8
	//mov good = 9,10
	//diode good = 11,12
	//cap good = 13
	//resistor good 14,15
	
	// figure out what components are good and bad; change graphics accordingly	

	#define component_y 28
	if (status&0x80)
		{
		vram[(component_y*32)+0]=138+9;
		vram[(component_y*32)+1]=138+10;
		}
	else
		{
		vram[(component_y*32)+0]=138+0;
		vram[(component_y*32)+1]=138+1;
		}

	if (status&0x40)
		{
		vram[(component_y*32)+3]=138+11;
		vram[(component_y*32)+4]=138+12;
		}
	else
		{
		vram[(component_y*32)+3]=138+3;
		vram[(component_y*32)+4]=138+4;
		}

	if (status&0x20)
		{
		vram[(component_y*32)+7]=138+13;
		}
	else
		{
		vram[(component_y*32)+7]=138+5;
		}

		
	if (status&0x10)
		{
		vram[(component_y*32)+10]=138+14;
		vram[(component_y*32)+11]=138+15;
		}
	else
		{
		vram[(component_y*32)+10]=138+6;
		vram[(component_y*32)+11]=138+7;
		}

	if (status&0x08)
		{
		vram[(component_y*32)+14]=138+13;
		}
	else
		{
		vram[(component_y*32)+14]=138+5;
		}

	if (status&0x04)
		{
		vram[(component_y*32)+17]=138+11;
		vram[(component_y*32)+18]=138+12;
		}
	else
		{
		vram[(component_y*32)+17]=138+3;
		vram[(component_y*32)+18]=138+4;
		}

	if (status&0x02)
		{
		vram[(component_y*32)+20]=138+9;
		vram[(component_y*32)+21]=138+10;
		}
	else
		{
		vram[(component_y*32)+20]=138+0;
		vram[(component_y*32)+21]=138+1;
		}
}

void collision_checks()
{
	unsigned char i,j;
	char mx,my,sx,sy;

	for (i=0;i<8;i++)
		{
		if (missiles[i].state>=1)  // if missile is active
			for (j=0;j<16;j++)
				{
				if (sparks[i].state!=0) // if spark is active
					{
					mx=(missiles[i].x>>8)+3; // move some calculations out here to be easier to read
					my=(missiles[i].y>>8)+4; // center of missile
					sx=(sparks[j].x>>8)-3; // and a little to the left of the spark
					sy=(sparks[j].y>>8)-4; // and a little to the top of the spark
					if ((mx>=sx)&&(mx<=(sx+9))) // then a little extra on the right
						if ((my>=sy)&&(my<=(sy+12))) // and a little extra on the bottom (basically making a grazing hit count)
							{
							sparks[j].state=0; // if we're in here, we hit.  Turn off spark
							missiles[i].state=2; // set the missile to start exploding
							missiles[i].x=sparks[j].x; // set the missile to the sparks old position
							missiles[i].y=sparks[j].y;
							sparks[j].x=-1; // and hide the spark sprite offscreen
							TriggerFx(1,0xa0,true); // play the 'hit spark' soundFX
							}
					}
				}
		}
}



void update_sparks()
{
	unsigned char i,j; // loop counter
	unsigned char temp;
	unsigned char sx;
	unsigned char left,right;

	for (i=0;i<16;i++)  // go through all the sparks possible
		{
		if (sparks[i].state>0)
			{
				sparks[i].x+=sparks[i].xstep; // add the motion vectors
				sparks[i].y+=sparks[i].ystep;

				// test for PCB impact
				if ((sparks[i].y>>8)>200) // too far down on the screen
					{
					for (j=0;j<7;j++)
						{
						sx=(sparks[i].x>>8); // this is the left edge of a spark
						right=pgm_read_byte(&(hit_locations2[j]));  // this is the span that constitutes a 'hit'
						left=pgm_read_byte(&(hit_locations1[j]));

						if (((sx<=right) && (sx>=left)) || (((sx+5)<=right+3) && ((sx+5)>=left+3)))
							{
							temp=pgm_read_byte(&(hit_masks[j]));
							temp^=0xff;
							bases&=temp;

							if (bases!=bases_old)
								{
								start_explosion(temp^0xff);
								bases_old=bases;
								}
							break;
							}
						}
					sparks[i].state=0;
					sparks[i].x=-1;
					
					}

				// test for wall impact
				if ((sparks[i].x>>8)<6)
					{
					sparks[i].xstep=-sparks[i].xstep; // bounce off
					}
				if ((sparks[i].x>>8)>130)
					{
					sparks[i].xstep=-sparks[i].xstep;
					}

				if((sync_count&0x03)==0) // once every four frames, so some animation
					{
					sparks[i].state++; // using the 'state' for the animation counter as well
					if (sparks[i].state==20+12) // it we incremented into the end of the animation loop
						{
						sparks[i].state=20; // reset to the beginning of the animation
//						sparks[i].anim_timer=50+rand()%50; 
						}
					}
			}
				
		sprites[10+i].x=(sparks[i].x)>>8; // the shift is to just use the MSB of the 8:8 fixed point value
		sprites[10+i].y=(sparks[i].y)>>8;
		sprites[10+i].tileIndex=sparks[i].state; // this does animation 
		}
}	

void update_missiles()
{
	unsigned char i; // loop counter


	for (i=0;i<8;i++)  // go through all the missiles
		{
		switch (missiles[i].state)
		{ 
		case 1: // missile is 'in the air'
			{
			if (missiles[i].range>0)  // if it still has range...
				{
				missiles[i].range--;  // reduce the range a bit
				missiles[i].x+=missiles[i].xstep; // add the motion vectors
				missiles[i].y+=missiles[i].ystep;
				}
			else
				{
				missiles[i].state=0;  // if it's out of range, change states
				missiles[i].x=-1;
				missiles[i].y=-1;
				}
			break;
			}

		case 2: // yeah, this is dumb, but since I had the switch/case here I was too tired and lazy to change it
		case 3: // when I added the spark exploding animation
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
		case 12:
		case 13:
		case 14:
		case 15:
		case 16:
		case 17:
		case 18:
		case 19:
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30: // 'exploding' states
			{
			missiles[i].state++;  // state value is used to point to explosion frame in sprite list too
			break;
			}
		case 31:
			{
			missiles[i].state=0;  // done exploding, hide sprite
			missiles[i].x=-1;
			missiles[i].y=-1;
			}
		}
				
		sprites[i+2].x=(missiles[i].x)>>8; // the shift is to just use the MSB of the 8:8 fixed point value
		sprites[i+2].y=(missiles[i].y)>>8;
		if (missiles[i].state==1)
			sprites[i+2].tileIndex=missiles[i].state; // this does animation if there is any on the bullets (unimplemented)
		if (missiles[i].state>1)
			{
			sprites[i+2].tileIndex=31+(missiles[i].state>>3); // this does animation when exploding too
			}

		}

	if (reload_time>0) // good of place as any-- time until next missile is available to fire timeout
		reload_time--;
}	

void fire_missile(char side)
{
	unsigned char i;   // loop counter
	int rise; // distance from firing base
	int run;
	int m;    // slope of resulting line to plot missile vector
	int n;

	for (i=0;i<8;i++)
		if (missiles[i].state==0)
			{
			missiles[i].state=1; //1== firing
			missiles[i].y=((8*24)+4)<<8; // 8 scanlines per row, 24 rows down for MOV.  Shifted to upper byte.
			rise=(unsigned int)((196-player_y)); // rise is MOV Y position - cursor (0=top of screen, 208 = bottom)
			if (side==0)
				{
				missiles[i].x=8<<8;  // starting position is position of MOV, shifted into upper byte for 8:8 fixed pt.
				run= (unsigned int)((player_x-8));   // run is cursor position minus MOV position (0=left, 132= right)
				n=(run<<8)/rise; // multiply run by 256, then divide by rise to avoid FP math (this is fixed point)
				}
			else
				{
				missiles[i].x=128<<8;  // starting position is position of MOV, shifted into upper byte for 8:8 fixed pt.
				run= (unsigned int)((128-player_x));   // run is cursor position minus MOV position (0=left, 132= right)
				n=(run<<8)/-rise; // multiply run by 256, then divide by rise to avoid FP math (this is fixed point)
				}

			m=(rise<<8)/run; // shift rise up to upper byte, then divide to avoid floating point
				
			// the non-square pixels make for faster horizontal fire than vertical... Not real sure how to 
			// compensate for that--> it's a feature.
			if ((rise)>run)			// decide which quadrant the vectors need to go to
				{
				missiles[i].ystep=-1*256;
				missiles[i].xstep=(int)n;
				missiles[i].range=rise;
				}
			else
				{
				if (side==0)		// invert xstep depending on which MOV is firing
					missiles[i].xstep=1*256;
				else
					missiles[i].xstep=-1*256;

				missiles[i].ystep=-(int)(m);
				missiles[i].range=run;
				}

			reload_time=30; // number of frames before next missile available.	

			TriggerFx(2,0x70,true); // missile firing sound effect

			break; // missile allocated, get out of here
			}
}

void new_spark()
{
	unsigned char i;   // loop counter

	for (i=0;i<16;i++)
		if (sparks[i].state==0)
			{
			sparks[i].state=20; //20== active, start of spark shapes
			sparks[i].x=(148-sx3)<<8;  // x position based on finger scroll offset
			sparks[i].y=(80-screenSections[0].scrollY)<<8; // adjust for tip of finger and Y scroll offset
			sparks[i].xstep=(int)(rand()%50);
			sparks[i].ystep=(int)((rand()%50)+40);
			if (wave>=15)
				sparks[i].ystep=(int)((rand()%50)+60);
			if (rand()%2)
				sparks[i].xstep*=-1;
			
			sparks[i].anim_timer=50+rand()%100; // number of frames before next animation cycle

			TriggerNote(3,5,35,0xff); // zap

			break; // spark allocated, get out of here
			}

}

void screen_setup()
{
	unsigned char i;

 	//Define the screen sections. 

	screenSections[0].height=80;	
	screenSections[0].flags=SCT_PRIORITY_SPR;

	screenSections[1].height=96;
	screenSections[1].vramBaseAdress=vram+(32*14);
	screenSections[1].flags=SCT_PRIORITY_SPR;

	screenSections[2].height=1;
	screenSections[2].vramBaseAdress=vram+(32*(14+12));
	screenSections[2].flags=SCT_PRIORITY_SPR;

	screenSections[3].height=1;
	screenSections[3].vramBaseAdress=vram+(32*(14+12));
	screenSections[3].flags=SCT_PRIORITY_SPR;
	screenSections[3].scrollY=1;

	screenSections[4].height=1;
	screenSections[4].vramBaseAdress=vram+(32*(14+12));
	screenSections[4].flags=SCT_PRIORITY_SPR;
	screenSections[4].scrollY=2;

	screenSections[5].height=1;
	screenSections[5].vramBaseAdress=vram+(32*(14+12));
	screenSections[5].flags=SCT_PRIORITY_SPR;
	screenSections[5].scrollY=3;

	screenSections[6].height=1;
	screenSections[6].vramBaseAdress=vram+(32*(14+12));
	screenSections[6].flags=SCT_PRIORITY_SPR;
	screenSections[6].scrollY=4;

	screenSections[7].height=1;
	screenSections[7].vramBaseAdress=vram+(32*(14+12));
	screenSections[7].flags=SCT_PRIORITY_SPR;
	screenSections[7].scrollY=5;

	screenSections[8].height=1;
	screenSections[8].vramBaseAdress=vram+(32*(14+12));
	screenSections[8].flags=SCT_PRIORITY_SPR;
	screenSections[8].scrollY=6;

	screenSections[9].height=1;
	screenSections[9].vramBaseAdress=vram+(32*(14+12));
	screenSections[9].flags=SCT_PRIORITY_SPR;
	screenSections[9].scrollY=7;

	screenSections[10].height=1;
	screenSections[10].vramBaseAdress=vram+(32*(14+12+1));
	screenSections[10].flags=SCT_PRIORITY_SPR;

	screenSections[11].height=1;
	screenSections[11].vramBaseAdress=vram+(32*(14+12+1));
	screenSections[11].flags=SCT_PRIORITY_SPR;
	screenSections[11].scrollY=1;

	screenSections[12].height=2;
	screenSections[12].vramBaseAdress=vram+(32*(14+12+1));
	screenSections[12].flags=SCT_PRIORITY_SPR;
	screenSections[12].scrollY=2;

	screenSections[13].height=4;
	screenSections[13].vramBaseAdress=vram+(32*(14+12+1));
	screenSections[13].flags=SCT_PRIORITY_SPR;
	screenSections[13].scrollY=4;

	screenSections[14].height=16;
	screenSections[14].vramBaseAdress=vram+(32*(14+12+2));
	screenSections[14].flags=SCT_PRIORITY_SPR;

	screenSections[0].tileTableAdress=sparkfun;
	screenSections[1].tileTableAdress=sparkfun;
	for (i=2;i<14;i++)
		screenSections[i].tileTableAdress=smokeframes2;
	screenSections[14].tileTableAdress=sparkfun;
}

int main(){	

	unsigned char i,j=0,scroller,char_index;
	unsigned char blink_rate=0;    // flash rate for "protect components!" message
	unsigned char timer1=0;		   // timer for MOV cooldown
	unsigned char exp_rate=0;	   // speed of explosion animation
	unsigned char smoke_frame=0;   // animation counter for smoke animation
	unsigned char wind_offsets[12];// used to offset sine table for wind effect
	unsigned char wind_counter=0;  // used to index into wind_offsets table
	unsigned char wind_mod1;	   // relative wind strength by height
	unsigned char wind_mod2;	   // randomized, but adjusted to top = heaviest, bottom-lowest
	unsigned char wind_mod3;
	unsigned char finger_rate=0;   // counter to set speed across top of screen
	unsigned char level_speed=1;   // match value to finger_rate to start motion
	unsigned char spark_time=200;  // time before sparks start from beginning of game
	unsigned char instruct_flag=0; // turns one after initial entry to instructions routine
	unsigned char scroll_done=0;
	unsigned char game_over_timeout=200;
	unsigned char wave_counter=0;
	
	unsigned char spark_speed=50;

	ClearVram(); // clear screen
		
	DDRD |= 0x10;  // PD3 = power led as output
  	PORTD |= 0x08; // power button input enabled, pulled up

	//initialize Alec's audio stuff
	
	InitMusicPlayer(patches);
	SetMasterVolume(0xd0); //crank up the volume a bit since the sample is not too loud

	SetSpritesTileTable(Invaders);  // where the sprites start
	SetSpritesOptions(SPR_OVERFLOW_CLIP);
	
	screen_setup(); // setup display list for scrolling (or not) regions

	//mov blown = 90+48+0,1
	// blank sprite =2
	//diode blown = 90+48+3,4
	//cap blown = 90+48+5
	//resistor blown = 90+48+6,7 
    // PCB FR4 = 90+48+8
	//mov good = 9,10
	//diode good = 11,12
	//cap good = 13
	//resistor good 14,15

	Fill(0,11,22,4,0); // clear background below finger in top scroll region (but above SparkFun logo in memory)
	SetFontTilesIndex(90+48+16);

	Print(1,16,strTitle);
	Print(1,17,strRev);
	Print(1,19,strCopy);
	TriggerNote(3,5,35,0xff); // zap

	WaitVsync(180);

	Fill(0,11,22,4,0); // clear background below finger in top scroll region (but above SparkFun logo in memory)
	Fill(0,25+4,22,1,90+48+8); // clear background below SparkFun logo

	DrawMap2(1,14,sparkfun_sparkfun_map); // draw SparkFun logo
	DrawMap2(23,0,finger_FingerMap); // draw hand/finger
	update_base(bases); // draw bases

	Print(1,21+3,strProtect); // 'protect components!' message

	for (i=0;i<12;i++)		// offsets for the wind sine lookup table (0,1,2,3,4,...,11)
		wind_offsets[i]=i;

	wind_mod1=4;	// default wind setting to blow smoke around
	wind_mod2=2;
	wind_mod3=1;

	player_x=72;	// starting cursor position
	player_y=100;
	sprites[1].tileIndex=0; // aiming cursor

	sprites[27].x=9;
	sprites[27].y=208;
	sprites[27].tileIndex=37; // MOV temperature meters to 'zero' image

	sprites[28].x=7+122;
	sprites[28].y=208;
	sprites[28].tileIndex=37;

	char_index=0;
	scroller=0;

	//////////////////////////////////////////////////////////////////
	// MAIN LOOP

	while(1){

		// this is just a little sample to show how the power button and LED can be used
		// it just blanks the screen and turns the LED on/off, but you could sleep the AVR
		// and use a pin change interrupt to wake it again, etc.

		if ((PIND&0x08)==0) // if power button is pressed on baseboard
			{
			DDRB &= 0xF7; // turn off video encoder clock
			WaitVsync(15); // wait 250ms to debouce button
			while ((PIND&0x08)!=0) // then as long as the button isn't pushed
				{
				PORTD |= 0x10;  // turn the LED off (off, off, off, off!)
				}
			while ((PIND&0x08)==0); // the button *was* pressed again, now wait until they release it (ghetto debouce)
			PORTD &= 0xEF; // power LED back on
			DDRB |= 0x08;  // encoder clock back on.
			}

		if (game_start==0)
			{
			blink_rate++;
			if (blink_rate>=90)
				blink_rate=0;
			if (blink_rate<=45)
				Print(1,21+3,strProtect);
			else
				Fill(0,21+3,22,1,0);
			}

		// this is what happens when you make stuff up as you go along.  Spaghetti code follows.
		// "hey, I should let you start a new game, and restart the instruction scroller, and show the max wave, and..."

		if (game_start==2) // game over
			{
			blink_rate++;
			if (blink_rate>=90)
				blink_rate=0;
			if (blink_rate<=45)
				Print(0,21+3,strGameover); // flash the game over message
			else
				Fill(0,21+3,22,1,0);  // (by overwriting it with black spaces now and then)

			Print(0,25,strHigh);   // we got here when a game ended, so print the max wave for bragging rights.
			PrintByte(17,25,wave);


			if (game_over_timeout==0) // game_over_timeout is just a shield time to keep this from happening right away.
				{
				bases=0xFE;			 // restore bases to living status
				Fill(0,21+5,22,2,0); // clear background below SparkFun logo  (erase smoke)
				game_start=0;		 // and the game is not started
				for (i=0;i<16;i++)	 // hide any sparks still on screen
					{
					sparks[i].state=0;
					sparks[i].x=140<<8;
					sparks[i].y=208<<8;
					}
				update_sparks();
				for (i=0;i<8;i++)    // and hide any missiles too
					{
					missiles[i].state=31;
					}
				update_missiles();
				while(sx3!=0)        // quickly scroll the hand off screen if it's on screen
					{
					WaitVsync(1);
					sx3++;
					if(sx3>X_SCROLL_WRAP) 
						sx3=0;		
					screenSections[0].scrollX=sx3;		
					}
				sx3+=8;				 // adjust scroll position to text draw is off-screen
				screenSections[0].scrollX=sx3;		
				instructions=1;      // instruction mode back on
				screenSections[0].tileTableAdress=SF_font-(32*6*8); // set the tileset to the text font instead of smoke graphics
				screenSections[0].scrollY=0;	// undo any vertical scrolling
				Fill(0,0,32,10,32); // clear background on top of screen
				WaitVsync(1); // and let the screen redraw with these changes
				j=0; // used by text scroller
				wave_counter=0; // and used by the game
				scroll_done=0; // and by the text scroller
				scroller=0; // and by the text scroller
				char_index=0; // and by the text scroller (can you tell I wasn't planning on reusing the scroller after the first run?)
				level_speed=2; // and set the level speed to normal
				wave=0; // and restart the waves at zero
				spark_speed=50; // and set the spark emergence time to normal
				
				WaitVsync(120); // wait a couple seconds so they can admire their wave achievement
				Print(0,25,strBlank); // then clear it from the screen
				}

			}

		//syncronize gameplay on vsync (30 hz)	
		WaitVsync(1);
		sync_count++; // counters used for various timing events below
		exp_rate++;
		wind_counter++;
		finger_rate++;

		// we do three passes here with smaller 'step' values per pass
		// this gives us collision detection between the steps of movement
		// seen on the display-- hence no missiles flying through sparks, etc.
		for (i=0;i<3;i++)
			{
			update_missiles();
			update_sparks();
			collision_checks();
			}


		if (((bases&0x7C)==0)&&(game_start==1)) // all components blown up (MOVs don't count)
			{
			game_start=2; // game over
			game_over_timeout=200;
			}

		if ((game_start==2)&&(sx3==0))  // if a game is recently over, wait until the text has scolled off screen
			instructions=0;				// ...and then turn the instructions off
			
		if (exp_rate==3)				// ever four frames
			{
			process_explosions();
			exp_rate=0;

			// seven frames in the smoke animation
			if (++smoke_frame>=7)
				smoke_frame=0;

			// if a base is blown up, draw and animate a column of smoke above it
			if ((bases^0xff)&0x02)
				DrawMap2(20,21+5,smoke_anim+(smoke_frame*6));
			if ((bases^0xff)&0x04)
				DrawMap2(17,21+5,smoke_anim+(smoke_frame*6));
			if ((bases^0xff)&0x10)
				DrawMap2(10,21+5,smoke_anim+(smoke_frame*6));
			if ((bases^0xff)&0x40)
				DrawMap2(3,21+5,smoke_anim+(smoke_frame*6));
			if ((bases^0xff)&0x80)
				DrawMap2(0,21+5,smoke_anim+(smoke_frame*6));

			update_base(bases);
			}

		timer1++;		// this is just sort-of a slow time base
		if (timer1==30)
			{
			if (temp_left>0) // if MOV temperatures are up, cool them one step
				temp_left--;
			if (temp_right>0)
				temp_right--;

			if (temp_left>4) // is the MOV too hot?
				sprites[27].tileIndex=42+(temp_left&0x01);  // yes, go to the red warning tiles and let it alternate
			else
				sprites[27].tileIndex=37+temp_left; // no, just show the temperature reading

			if (temp_right>4) // (same for other MOV)
				sprites[28].tileIndex=42+(temp_right&0x01);
			else
				sprites[28].tileIndex=37+temp_right;

			timer1=0;

			if (++wave_counter>20) // count the number of these slow timer ticks to use for wave progression
				{
				wave++; 		// been a while, go to new wave
				wave_counter=0; // reset count
				Print(8,24,strWave); // print the wave number message
				PrintByte(14,24,wave); // and the wave number next to it
				spark_speed-=2; // speed up the sparks a bit
				if (spark_speed<5) // but not *too* fast-- keep it moderately impossible
					spark_speed=5;
				}

//			if (wave==15)  // this was a trigger to speed up/slow down the hand, but it seemed too 'busy' with it in
//				{
//				level_speed=2;
//				finger_rate=0;
//				}
			}

	if (wave_counter==5)  // if the wave counter has been going for a bit...
		Print(8,24,strBlank); // clear the "wave x" message

	if ((sync_count&0x2f)==0x2f) // every once in a while
		{
		if ((rand()%2)==0) // and with a little randomness
			{
			wind_mod1++;
			if (wind_mod1>5)  // change the windspeed on the smoke up
				wind_mod1=5;
			}
		else
			{
			wind_mod1--;	  // ...or down.
			if (wind_mod1<3)
				wind_mod1=3;
			}				

		wind_mod2=wind_mod1-2;	// make lower portions of the smoke move slower		
		wind_mod3=wind_mod2-1;  // yeah, I got a little obsessive with the smoke behaviour
		}

	for (i=2;i<12;i++)  // apply the smoke motion by line scrolls
		{
		screenSections[i].scrollX=pgm_read_byte(&(windy[wind_offsets[i-2]]))-wind_mod1;				
		}
		screenSections[12].scrollX=(pgm_read_byte(&(windy[wind_offsets[i-2]]))>>1)-wind_mod2;				
		screenSections[13].scrollX=(pgm_read_byte(&(windy[wind_offsets[i-2]]))>>2)-wind_mod3;				

	if (wind_counter==5) // step through the offsets in the ram table so the next table lookup is progressive on the sine table
		{
		wind_counter=0;
		for (i=0;i<12;i++)
			{
			wind_offsets[i]++;
			if (wind_offsets[i]>=12)
				wind_offsets[i]=0;
			}
		}

	if (instructions==0) // if the instructions aren't running, run the game behaviour
		{
		level_speed=2;   // speed control for the hand/text scroller.  Hand always runs at '2' now.
		if (finger_rate==level_speed)
			{
			finger_rate=0;
			sx3++;
			if(sx3>X_SCROLL_WRAP) sx3=0;  // scroll the top region and apply a sinus to the Y		
	
			screenSections[0].scrollX=sx3;				
			screenSections[0].scrollY=pgm_read_byte(&(finger_sine[(sx3&0x3f)]));	
			}
		}
	else	// not in the game, so we're a text scroller!
		{
		if (finger_rate==level_speed)
			{
			finger_rate=0; // still using this, just for the text speed
			level_speed=1;
			if (instruct_flag==0)
				{
				sx3+=8;    // shift the scroll a little so that new character plotting is invisible (offscreen)
				screenSections[0].tileTableAdress=SF_font-(32*6*8); // point graphics for that region to font instead of hand graphics
				Fill(0,0,32,10,32); // clear background on top of screen
				instruct_flag=1; // and officially in instructions mode
				j=0;
				}
			
			if(scroll_done==0)
				{
				scroller++;
				if (scroller>5) // scroller counts six times-- six pixels in each character
					{
					vram[(32*5)+(++j)]=pgm_read_byte(&(strScroller[char_index++]));  // then draws the next character from the message into memory
					scroller=0; // and gets ready to do it again, and again, and again...
					}
				// ok, this is awful, but basically adjust the memory pointer back after every 32 byte so that it stays
				// on the same line of text.  Could use a counter, or maybe % or something.  I kept adding on to the
				// message and before long the 'if' statement kept growing.  The AVR's so fast you can be sloppy. :-)
				if ((char_index==31)||(char_index==63)||(char_index==95)||(char_index==127)||(char_index==159)||(char_index==191)||(char_index==223))
					j=255;
				// if we hit the end of the text message...
				if (char_index>(sizeof(strScroller)-2))
					{
					char_index=0; // reset to beginning
					scroll_done=1;// say we're done
					j=0; // set our destination memory pointer back just in case.
					}
				}
			sx3++;
			if(sx3>=X_SCROLL_WRAP) 
				sx3=0;		// take care of wrap, and do the scroll (finally)
			screenSections[0].scrollX=sx3;		
				
			if ((scroll_done==1)&&(sx3==0))		// and if we're done and the screen has scrolled into a safe spot...
				{
				instructions=0;					// back to game mode
				screenSections[0].tileTableAdress=sparkfun; // restore graphics to the hand graphic
				Fill(0,0,32,10,0); // clear the screen area
				DrawMap2(23,0,finger_FingerMap); // draw hand/finger
				}

			}
		}

	if (instructions==0) // again, only if the game mode is active
		{
		spark_time--;  // decrement the "time to appear" count for new sparks
//		PrintHexByte(2,21,sx3);
		if ( (spark_time==0))//&&(sx3>0x10)&&(sx3<0x80) )  // and if it's zero...
			{
			new_spark();  // spawn a new spark
			spark_time=spark_speed+rand()%spark_speed; // with a randomized reload time based on the difficulty level
			}
		}

	sprites[1].x=player_x;  // position aiming cursor
	sprites[1].y=player_y;

	if (game_over_timeout==0) // lock out controls to make them see the scrolled and messages
		processControls();
	else
		game_over_timeout--;
	}
}

unsigned char processControls(void){
	static int lastbuttons=0;
	unsigned int joy=ReadJoypad(0);



	if((joy&BTN_START) && !(lastbuttons&BTN_START)){
	}


	if(joy&BTN_SR){
	}
	
	if(joy&BTN_SL){
	}


	if(joy&BTN_B){
		if (instructions==1) // kind of a hack way to start the game...
			{
				instructions=0; // exit the instruction scroller
				screenSections[0].tileTableAdress=sparkfun; // put tile pointer back to 'hand' graphics
				Fill(0,0,32,10,0); // clear the background
				sx3=0; // set scrolling area offscreen for redraw
				screenSections[0].scrollX=sx3;	// do move	
				DrawMap2(23,0,finger_FingerMap); // draw hand/finger
			}
			
	
		if (reload_time==0)
			{
			if (temp_left>4) // firing when the MOV is too hot?
				{
				temp_left=10;  // tsk, tsk... penalty time!
				TriggerFx(0,0xa0,true);  // make annoying noise at the offender
				}
			else
				{
				temp_left++;  // firing warms up the MOV
				if (bases&0x80)  // is the left MOV blown up?
					fire_missile(0); // nope, shoot!
				else
					{
					if((rand()%3)==1) // blown up, but might still be good for a shot once in a while
						fire_missile(0);
					}
				}
			if (game_start==0) // and this will start a game if one's not in progress
				{
				wave=0; // reset to wave zero (probably not the only place I did this-- again, planning was not high on my todo list for this.)
				game_start=1; // start the game
				Fill(1,21+3,22,1,0); // clear the attract mode message
				}
			}
	}

	if(joy&BTN_A){
		if (instructions==1) // kinda of a hack way to start the game...
			{
				instructions=0; // exit the instruction scroller
				screenSections[0].tileTableAdress=sparkfun; // put tile pointer back to 'hand' graphics
				Fill(0,0,32,10,0); // clear the background
				sx3=0; // set scrolling area offscreen for redraw
				screenSections[0].scrollX=sx3;	// do move	
				DrawMap2(23,0,finger_FingerMap); // draw hand/finger
			}

		if (reload_time==0)
			{
			if (temp_right>4)
				{
				temp_right=10;
				TriggerFx(0,0xa0,true);
				}
			else
				{
				temp_right++;
				if (bases&0x02)
					fire_missile(1);
				else
					{
					if((rand()%3)==1)
						fire_missile(1);
					}
				}
			if (game_start==0)
				{
				wave=0;
				game_start=1;
				Fill(1,21+3,22,1,0);
				}
			}
	}



	// check for player moving the cursor
	// add values as appropriate, and constrain movement to playfield
	if (joy&BTN_UP)
	{
	player_y-=2;		// vertical axis moves at 2x the horizontal to compensate for rectangular pixels
	if (player_y<=8)
		player_y=8;
	}
	if (joy&BTN_DOWN)
	{
	player_y+=2;
	if (player_y>=192)
		player_y=192;
	}
	if (joy&BTN_LEFT)
	{
	player_x--;
	if (player_x<=6)
		player_x=6;

	}
	if (joy&BTN_RIGHT)
	{
	player_x++;
	if (player_x>132)
		player_x=132;
	}


	
	if(joy&BTN_X){
	}

	if(joy&BTN_Y){
	}
		
	if(joy&BTN_SELECT){
	}

	lastbuttons=joy;

	return 0;
}

